from dataclasses import dataclass

from goodwiki.mwp_utils import replace_obj, remove_obj


@dataclass
class TagTranscludeStatus:
    tag_name: str  # name of tag
    unknown: bool  # whether the tag was not recognized


class TagTranscluder:
    def __init__(self, with_styling: bool):
        self.with_styling = with_styling
        self.tags_to_keep = {
            "br",
            "sub",
            "sup",
            "blockquote",
            "math",
            "syntaxhighlight",
            "code",
            "pre",
            "ul",
            "li",
            "ol",
            "nowiki",
        }
        if with_styling:
            self.tags_to_keep.update(["u", "del"])

        self.tags_to_extract = {
            "p",
            "abbr",
            "bdi",
            "bdo",
            "cite",
            "data",
            "dfn",
            "mark",  # not used frequently enough to warrant preservation in output
            "ruby",
            "rp",
            "rt",
            "rb",
            "samp",
            "small",
            "time",
            "var",
            "center",
            "font",
            "a",
            "big",
            "onlyinclude",
            "section",
            "chem",
            "ce",
        }
        if not with_styling:
            self.tags_to_extract.update(
                [
                    "u",
                    "b",
                    "strong",
                    "i",
                    "em",
                    "del",
                    "s",
                    "strike",
                    "''",
                    "'''",
                    "'''''",
                ]
            )

        self.tags_to_delete = {
            "style",
            "hr",
            "wbr",
            "div",
            "table",
            "td",
            "tr",
            "th",
            "caption",
            "thead",
            "tfoot",
            "tbody",
            "noinclude",
            "hiero",
            "ref",
            "templatestyles",
            "phonos",
            "gallery",
            "indicator",
            "timeline",
            "references",
            "img",
        }

    def parse_attributes(self, raw_attr: list[str]) -> dict[str, str | None]:
        attributes = {}
        for a in raw_attr:
            a = a.strip()
            comps = a.split("=")
            prop = comps[0]
            val = None
            if len(comps) == 2:
                val = comps[1]
                if prop == "class":
                    val = val.strip()
            attributes[prop] = val
        return attributes

    def transclude(self, tag_obj, section) -> TagTranscludeStatus:
        tag_name = tag_obj.tag.strip().lower()
        tag_attrs = self.parse_attributes(tag_obj.attributes)
        unknown = False
        match tag_name:
            # Keep some tags as is
            case t if t in self.tags_to_keep:
                pass
            # Extract content from some tags
            case t if t in self.tags_to_extract:
                replace_obj(tag_obj, tag_obj.contents, section)
            # Delete some tags entirely, including content
            case t if t in self.tags_to_delete:
                remove_obj(tag_obj, section)
            # Rest of cases handle special tags (h1, b, i, em, ins, etc.)
            # if not handled already (like for no styling case)
            case "span":
                # flagicons have extra space that we need to remove
                if "flagicon" in tag_attrs:
                    remove_obj(tag_obj, section)
                else:
                    replace_obj(tag_obj, tag_obj.contents, section)
            case "h1" | "h2" | "h3" | "h4" | "h5" | "h6":
                # Replace with heading
                level = int(tag_name[1])
                equals_signs = "=" * level
                replace_obj(
                    tag_obj,
                    f"{equals_signs} {tag_obj.contents} {equals_signs}",
                    section,
                )
            case "b" | "strong":
                replace_obj(tag_obj, f"'''{tag_obj.contents}'''", section)
            case "i" | "em":
                replace_obj(tag_obj, f"''{tag_obj.contents}''", section)
            case "ins":
                # Replace with u
                replace_obj(tag_obj, f"<u>{tag_obj.contents}</u>", section)
            case "q":
                # Replace with text
                replace_obj(tag_obj, f'"{tag_obj.contents}"', section)
            case "s" | "strike":
                # Replace with del
                replace_obj(tag_obj, f"<del>{tag_obj.contents}</del>", section)
            case "kbd" | "samp" | "tt":
                # Replace with code
                replace_obj(tag_obj, f"<code>{tag_obj.contents}</code>", section)
            case "dl":
                # Replace with list
                replace_obj(tag_obj, tag_obj.contents, section)
            case "dt":
                replace_obj(tag_obj, f"; {tag_obj.contents}", section)
            case "dd":
                replace_obj(tag_obj, f": {tag_obj.contents}", section)
            case "poem":
                tag_obj.tag = "blockquote"
            case _:
                # If none of the previous cases match, remove outright
                remove_obj(tag_obj, section)
                unknown = True

        return TagTranscludeStatus(tag_name=tag_name, unknown=unknown)



removed_sections = {
    "references",
    "works cited",
    "bibliography",
    "further reading",
    "published works",
    "external links",
    "notes",
    "notes and references",
    "further reading",
    "external links",
    "literature cited",
    "sources",
    "citations",
    "sources and further reading",
    "footnotes",
    "sources cited",
}

# Some templates are really large sections of the page and
# should be removed before continuing preprocessing to minimize
# processing error. Use startswith for detection
big_templates_to_remove = {
    "football",
    "fs ",  # shorthand for football squad
    "quote box",
    "infobox",
    "multiple",
    "portal",
    "listen",
    "speciesbox",
    "video game reviews",
    "tree chart",
    "album ratings",
    "film ratings",
    "game ratings",
    "music ratings",
    "song ratings",
    "television episode ratings",
    "television season ratings",
    "television ratings",
    "tv ratings",
    "wide image",
    "automatic taxobox",
    "pp",
    "pp-move",
    "pp-vandalism",
    "s-ttl",
}

# Template names should have the first letter lowercased, with
# underscores replaced by spaces
templates_to_keep = {
    "abbr",
    "asterisk",
    "as of",
    "bibleverse",
    # blockquote
    "blockquote",
    "bq",
    "quote",
    "bquote",
    "quotation",
    # birth date
    "birth date",
    "birth date and age",
    "birth year",
    "birth year and age",
    "death date",
    "death date and age",
    "death year and age",
    # end birth date
    "block indent",
    "center",
    "circa",
    # clear
    "clear",
    "-",
    "clr",
    # chem
    "chem",
    "chem2",
    # end chem
    "convert",
    "cquote",
    "currentyear",
    "cvt",
    "dash year",
    "date",
    "delta",
    "dinogloss",
    "em",
    # scientific notation
    "e",
    "10^",
    "x10^",
    # em dash
    "em dash",
    "--",
    "mdash",
    # en dash
    "endash",
    "en dash",
    "ndash",
    "nsndns",
    # end en dash
    "equationNote",
    "equationRef",
    "f1",
    # flag
    "flag",
    "flagu",
    "flagIOC",
    "flagIOC2",
    "flaglink",
    "fbw",
    "flagathlete",
    # format price
    "format price",
    "formatprice",
    # fraction
    "fraction",
    "frac",
    "sfrac",
    # end fraction
    "gRT",
    "hyphen",
    "ill",
    "interlanguage link",
    # inflation
    # "inflation", doesn't work well with expandtemplates
    # "inflation-year",
    # "inflation/year",
    # rupee
    "iNR",
    "indian Rupee",
    "iNRConvert",
    # phonetics
    "iPAc-en",
    "respell",
    # end phonetics
    "inrconvert",
    "jcon",
    "jct",
    "lang",
    "langnf",
    "lambda",
    # h-phrases
    "h-phrases",
    "p-phrases",
    # literal translation
    "literal translation",
    "lit",
    "lit.",
    "literally",
    # end literal translation
    "mabs",
    # sports years
    "mLB Year",
    "mlby",
    "baseball year",
    "nbay",
    "nBA Year",
    "nFL Year",
    # math
    "math",
    "mset",
    "mvar",
    "abs",
    "norm",
    # mono
    "mono",
    "monospace",
    "tt",
    "ttt",
    # nbsp
    "nbsp",
    "&nbsp",
    "nbs",
    "nbsp",
    "nbsp;",
    "space",
    "spcs",
    # non-breaking en dash
    "non breaking en dash",
    "nbnd",
    "nbndash",
    # end non-breaking en dash
    "navy",
    "nee",
    "nihongo",
    "nihongo2",
    "nFLplayer",
    "nobold",
    "nuclide",
    # not a typo
    "not a typo",
    "as written",
    "chem name",
    "proper name",
    "typo",
    "not typo",
    "nat",
    "NAT",
    # nostyling
    "nobold",
    "noitalic",
    "ni",
    "yesitalic",
    # end nostyling
    "nowrap",
    "nobr",
    "nobreak",
    # "numBlk",  TODO: don't remove div generated by this
    "okina",
    # percent
    "percent",
    "percentage",
    "pct",
    "winning percentage",
    # end percent
    "pi",
    "plainlist",
    "post-nominals",
    # poemquote
    "poem quote",
    "poemquote",
    # end poemquote
    "radic",
    "ratio",
    # refbegin
    # this is necessary to avoid contents rendering by accident in markdown
    # div HTML tag removal should get rid of this anyways
    "refbegin",
    "refend",
    # reign
    "reign",
    "r.",
    "reigned",
    "rexit",
    "ruled",
    # round
    "round",
    "rnd",
    # ships (TODO: incomplete)
    "sclass",
    "sclass2",
    "ship",
    "hMS",
    "sS",
    "uSNS",
    "uSS",
    "sMS",
    "sMU",
    # end ships
    "sic",
    "small",
    # station link
    "stl",
    "station link",
    "rws",
    "stnlnk",
    "station",
    "stn",
    # allcaps
    "allcaps",
    "caps",
    "nocaps",
    "smallcaps",
    "smallcaps2",
    "smallcapsall",
    "aut",
    "sc",
    "sc1",
    "sc2",
    "sm",
    # end allcaps
    "sms",
    # spaced en dash
    "spaced en dash",
    "spaced ndash",
    "dash",
    "snd",
    "spnd",
    "sndash",
    "spndash",
    # end spaced en dash
    "sqrt",
    "start date",
    # strikethrough
    "strikethrough",
    "strike",
    "s",
    # end strikethrough
    "strong",
    "sub",
    "sup",
    # transliteration
    "transliteration",
    "transl",
    "trans",
    # lists
    "unbulleted list",
    "ubl",
    "ubt",
    "ublist",
    "unbullet",
    "ordered list",
    "plain list",
    "plainlist",
    "pL",
    # end lists
    "ushr",
    # US politician abbreviation
    "uSpolabbr",
    "u.S. politician abbreviation",
    # end US politician abbreviation
    "uS$",
    "val",
    "wikidata",
    "yEAR",
    "zh",
    "+",
    "=",
    # bullet
    "•",
    "*",
    "bull",
    "bullet",
    # end bullet
    "'",
    "'-",
    "'s",
    "'\"",
    "\"'",
    # single + double
    "' \"",
    "single+double",
    # MAGIC VARIABLES
    "cURRENTYEAR",
    "cURRENTMONTH",
    "cURRENTMONTH1",
    "cURRENTMONTH2",
    "cURRENTMONTHNAME",
    "cURRENTMONTHNAMEGEN",
    "cURRENTMONTHABBREV",
}

parser_functions = {
    "#count",
    "#explode",
    "#expr",
    "#if",
    "#ifeq",
    "#iferror",
    "#ifexpr",
    "#ifexist",
    "#len",
    "#pos",
    "#replace",
    "#rpos",
    "#switch",
    "#time",
    "#timeel",
    "#urldecode",
    "formatnum",
    "#dateformat",
    "#formatdate",
}

_country_templates_raw = [
    "ABW",
    "ARU",
    "AFG",
    "AGO",
    "ANG",
    "ANT",
    "AHO",
    "AIA",
    "ALA",
    "ALB",
    "AND",
    "ARE",
    "UAE",
    "ARG",
    "ARM",
    "ASM",
    "ASA",
    "ATA",
    "ATF",
    "ATG",
    "AUS",
    "AUT",
    "AZE",
    "BDI",
    "BEL",
    "BEN",
    "BES",
    "BFA",
    "BGD",
    "BGR",
    "BUL",
    "BHR",
    "BHS",
    "BAH",
    "BIH",
    "BLM",
    "BLR",
    "BLZ",
    "BIZ",
    "BMU",
    "BER",
    "BOL",
    "BRA",
    "BRB",
    "BAR",
    "BRN",
    "BRU",
    "BTN",
    "BHU",
    "BVT",
    "BWA",
    "BOT",
    "CAF",
    "CTA",
    "CAN",
    "CCK",
    "CHE",
    "SUI",
    "SWI",
    "CHL",
    "CHI",
    "CHN",
    "PRC",
    "CHN-ROC",
    "CIV",
    "CMR",
    "COD",
    "COG",
    "CGO",
    "COK",
    "COL",
    "COM",
    "CPV",
    "CRI",
    "CRC",
    "CUB",
    "CUW",
    "CUR",
    "CXR",
    "CYM",
    "CAY",
    "CYP",
    "CZE",
    "CSK",
    "TCH",
    "DEU",
    "GER",
    "GDR",
    "DDR",
    "FRG",
    "DJI",
    "DMA",
    "DNK",
    "DEN",
    "DOM",
    "DZA",
    "ALG",
    "ECU",
    "EGY",
    "ERI",
    "ESH",
    "SADR",
    "ESP",
    "EST",
    "ETH",
    "FIN",
    "FJI",
    "FIJ",
    "FLK",
    "FRA",
    "FRO",
    "FSM",
    "GAB",
    "GBR",
    "UK",
    "ENG",
    "SCO",
    "SCT",
    "WAL",
    "WLS",
    "NIR",
    "GEO",
    "GGY",
    "GHA",
    "GIB",
    "GIN",
    "GUI",
    "GLP",
    "GMB",
    "GAM",
    "GNB",
    "GBS",
    "GNQ",
    "GEQ",
    "EQG",
    "GRC",
    "GRE",
    "GRD",
    "GRN",
    "GRL",
    "GTM",
    "GUA",
    "GUF",
    "GUM",
    "GUY",
    "HKG",
    "HKG-CHN",
    "HMD",
    "HND",
    "HON",
    "HRV",
    "CRO",
    "HTI",
    "HAI",
    "HUN",
    "IDN",
    "INA",
    "IMN",
    "IND",
    "IOT",
    "IRL",
    "IRN",
    "IRI",
    "IRQ",
    "ISL",
    "ISR",
    "ITA",
    "JAM",
    "JEY",
    "JOR",
    "JPN",
    "KAZ",
    "KEN",
    "KGZ",
    "KHM",
    "CAM",
    "KIR",
    "KNA",
    "SKN",
    "KOR",
    "ROK",
    "KWT",
    "KUW",
    "LAO",
    "LBN",
    "LIB",
    "LBR",
    "LBY",
    "LBA",
    "LCA",
    "LIE",
    "LKA",
    "SRI",
    "LSO",
    "LES",
    "LTU",
    "LUX",
    "LVA",
    "LAT",
    "MAC",
    "MAC-CHN",
    "MAF",
    "MAR",
    "MCO",
    "MON",
    "MDA",
    "MDG",
    "MAD",
    "MDV",
    "MEX",
    "MHL",
    "MKD",
    "NMK",
    "MLI",
    "MLT",
    "MMR",
    "MYA",
    "MNE",
    "MNG",
    "MGL",
    "MNP",
    "MOZ",
    "MRT",
    "MTN",
    "MSR",
    "MTQ",
    "MUS",
    "MRI",
    "MWI",
    "MAW",
    "MYS",
    "MAS",
    "MYT",
    "NAM",
    "NCL",
    "NER",
    "NIG",
    "NFK",
    "NGA",
    "NGR",
    "NIC",
    "NCA",
    "NIU",
    "NLD",
    "NED",
    "NOR",
    "NPL",
    "NEP",
    "NRU",
    "NZL",
    "NZ",
    "OMN",
    "OMA",
    "PAK",
    "PAN",
    "PCN",
    "PER",
    "PHL",
    "PHI",
    "PLW",
    "PNG",
    "POL",
    "PRI",
    "PUR",
    "PRK",
    "DPRK",
    "PRT",
    "POR",
    "PRY",
    "PAR",
    "PSE",
    "PLE",
    "PYF",
    "QAT",
    "REU",
    "ROU",
    "ROM",
    "RUS",
    "RWA",
    "SAU",
    "KSA",
    "SDN",
    "SUD",
    "SEN",
    "SGP",
    "SIN",
    "SGS",
    "SHN",
    "SJM",
    "SLB",
    "SOL",
    "SLE",
    "SLV",
    "ESA",
    "SMR",
    "SOM",
    "SPM",
    "SRB",
    "SSD",
    "STP",
    "SUR",
    "SVK",
    "SVN",
    "SLO",
    "SWE",
    "SWZ",
    "SXM",
    "SYC",
    "SEY",
    "SYR",
    "TCA",
    "TCD",
    "CHA",
    "TGO",
    "TOG",
    "THA",
    "TJK",
    "TKL",
    "TKM",
    "TLS",
    "ETM",
    "TON",
    "TGA",
    "TTO",
    "TRI",
    "TUN",
    "TUR",
    "TUV",
    "TWN",
    "ROC",
    "TWN-ROC",
    "ROC-TW",
    "TPE",
    "CHN-ROC",
    "TZA",
    "TAN",
    "UGA",
    "UKR",
    "UMI",
    "UNO",
    "URS",
    "URY",
    "URU",
    "USA",
    "US",
    "USSR",
    "UZB",
    "VAT",
    "VCT",
    "VIN",
    "VEN",
    "VGB",
    "IVB",
    "VIR",
    "ISV",
    "VNM",
    "VIE",
    "VUT",
    "VAN",
    "WLF",
    "WSM",
    "SAM",
    "YEM",
    "YUG",
    "FR-YUG",
    "ZAF",
    "RSA",
    "ZMB",
    "ZAM",
    "ZWE",
    "ZIM",
]
country_templates = {
    name.replace(name[0], name[0].lower(), 1).strip() for name in _country_templates_raw
}



import os

from collections import Counter
import multiprocessing as mp
from typing import Generator, cast

import mwparserfromhell as mwp
import pyarrow.parquet as pq
from tqdm import tqdm

from goodwiki.constants import removed_sections, big_templates_to_remove


def get_batch(filename: str, batch_size: int) -> Generator[list[dict], None, None]:
    """
    Get batch of given batch size from given parquet file.
    """
    pqfile = pq.ParquetFile(filename)
    for batch in pqfile.iter_batches(batch_size=batch_size):
        yield cast(list[dict], batch.to_pylist())


def count_rows(pq_filename: str, batch_size: int) -> int:
    """
    Count rows of parquet file of given filename.
    """
    return sum([len(batch) for batch in get_batch(pq_filename, batch_size)])


def _streamline_wikitext(wikitext: str) -> str:
    """
    Remove references, external links, notes, infoboxes, refs, and tables
    from wikitext, since they're already excluded from final output.

    Need to do this to more accurately estimate tag and template counts
    in dataset, since we're removing these sections/components anyways.
    """
    code = mwp.parse(wikitext)
    sections = []
    for section in code.get_sections(
        flat=True, include_lead=True, include_headings=True
    ):
        # Get rid of certain sections
        headings = section.filter_headings()
        heading = None
        if len(headings) != 0:
            heading = headings[0].title.strip().lower()
        if heading in removed_sections:
            continue

        # Get rid of tables and refs
        for obj in section.ifilter_tags(recursive=True):
            name = obj.tag.strip().lower()
            if name in {"table", "ref"}:
                try:
                    section.remove(obj)
                except ValueError:
                    pass

        # get rid of infoboxes
        for obj in section.ifilter_templates(recursive=True):
            name = obj.name.strip().lower()
            for temp_rmv in big_templates_to_remove:
                if name.startswith(temp_rmv):
                    try:
                        section.remove(obj)
                    except ValueError:
                        pass

        sections.append(str(section))
    return "\n\n".join(sections)


def combine_dicts(a: dict[str, set], b: dict[str, set]) -> dict[str, set]:
    return dict(
        list(a.items()) + list(b.items()) + [(k, a[k] | b[k]) for k in set(b) & set(a)]
    )


def _count_templates_task(chunk: list[dict]) -> tuple[Counter, dict[str, set]]:
    res = Counter()
    temp_counts_by_article = {}
    for c in chunk:
        article_template_counter = Counter()
        code = mwp.parse(_streamline_wikitext(c["wikitext"]))
        for temp in code.ifilter_templates(recursive=True):
            name = temp.name.strip().lower()
            article_template_counter[name] += 1
        res += article_template_counter
        temp_counts_by_article[c["title"]] = article_template_counter
    return res, temp_counts_by_article


def count_templates(pq_filename: str) -> tuple[Counter, dict[str, set]]:
    """
    Count template names used in wiki articles in the given
    parquet file.
    """
    num_cores = max(1, cast(int, os.cpu_count()) - 2)
    chunk_size = 128
    num_chunks = sum([1 for _ in get_batch(pq_filename, chunk_size)])
    counts = Counter()
    temp_counts_by_article = {}
    with mp.Pool(num_cores) as pool:
        for chunk_res in tqdm(
            pool.imap(_count_templates_task, get_batch(pq_filename, chunk_size)),
            total=num_chunks,
        ):
            counts += chunk_res[0]
            temp_counts_by_article.update(chunk_res[1])
    return counts, temp_counts_by_article


def _count_tags_task(chunk: list[dict]) -> tuple[Counter, dict[str, set]]:
    res = Counter()
    tag_counts_by_article = {}
    for c in chunk:
        article_tag_counter = Counter()
        code = mwp.parse(_streamline_wikitext(c["wikitext"]))
        for temp in code.ifilter_tags(recursive=True):
            name = temp.tag.strip().lower()
            article_tag_counter[name] += 1
        res = article_tag_counter
        tag_counts_by_article[c["title"]] = article_tag_counter
    return res, tag_counts_by_article


def count_tags(pq_filename: str) -> tuple[Counter, dict[str, set]]:
    """
    Count tag names used in wiki articles in the given
    parquet file. Note that tags are not just HTML tags, they
    also include wikitext equivalents for them like list items.

    This code is repetitive but I was lazy.
    """
    num_cores = max(1, cast(int, os.cpu_count()) - 2)
    chunk_size = 128
    num_chunks = sum([1 for _ in get_batch(pq_filename, chunk_size)])
    counts = Counter()
    tag_counts_by_article = {}
    with mp.Pool(num_cores) as pool:
        for chunk_res in tqdm(
            pool.imap(_count_tags_task, get_batch(pq_filename, chunk_size)),
            total=num_chunks,
        ):
            counts += chunk_res[0]
            tag_counts_by_article.update(chunk_res[1])
    return counts, tag_counts_by_article



import re

from goodwiki.constants import parser_functions


# Filters for magic words that are parser instructions -- e.g., __NOTOC__
RE_RM_MAGIC = re.compile("__[A-Z]*__", flags=re.UNICODE)
# Filters for file/image links.
_media_prefixes = "|".join(["File", "Image", "Media"])
RE_RM_WIKILINK = re.compile(
    f"^(?:{_media_prefixes}):", flags=re.IGNORECASE | re.UNICODE
)
# Leave category links in-place but remove the category prefixes
_cat_prefixes = "|".join(["Category"])
RE_CLEAN_WIKILINK = re.compile(
    f"^(?:{_cat_prefixes}):", flags=re.IGNORECASE | re.UNICODE
)


def try_replace_obj(obj, replacement, section):
    try:
        section.replace(obj, replacement)
    except ValueError:
        # For unknown reasons, objects are sometimes not found.
        # (mwparserfromhell issue)
        pass


def replace_obj(obj, replacement, section):
    section.replace(obj, replacement)


def try_remove_obj(obj, section):
    try:
        section.remove(obj)
    except ValueError:
        # For unknown reasons, objects are sometimes not found.
        # (mwparserfromhell issue)
        pass


def remove_obj(obj, section):
    section.remove(obj)


def get_heading(section) -> str | None:
    """Get heading from section"""
    headings = section.filter_headings()
    if len(headings) == 0:
        return None
    return headings[0].title.strip().lower()


def is_media_wikilink(obj):
    return bool(RE_RM_WIKILINK.match(str(obj.title)))


def is_category_wikilink(obj):
    return bool(RE_CLEAN_WIKILINK.match(str(obj.title)))


def tags_to_delete(obj):
    return str(obj.tag).lower() in {"ref", "table"}


def get_wikilink_text(wikilink_obj):
    """
    Get text (wikicode object) from wikilink, returning either the title field
    if text does not exist or the text field otherwise.
    """
    if wikilink_obj.text is not None and wikilink_obj.text != "":
        return wikilink_obj.text
    return wikilink_obj.title


def clean_wikilink(obj):
    text = get_wikilink_text(obj)
    text = re.sub(RE_CLEAN_WIKILINK, "", str(text))
    obj.text = text


def parse_params(template_obj) -> dict[str, str]:
    # Param names should be stripped, unlike values.
    # This is possible: {{tempname| arg=test}}
    # Params with empty values are not added
    params = {
        p.name.strip(): str(p.value) for p in template_obj.params if str(p.value) != ""
    }
    return params


def get_template_name(template_obj) -> str:
    """
    Get template name from template object, standardizing the name
    by replacing underscores with spaces and lowercasing the first letter.

    This matches the guidelines of template names on Wikipedia, which are
    case-sensitive except the first letter and have spaces that are equivalent
    to underscores. See: https://meta.wikimedia.org/wiki/Help:Template#Case_sensitivity

    Also strip the spaces on either side of the template name, as that's what
    Wikipedia implicitly allows.
    """
    raw_name = str(template_obj.name)
    name = raw_name.replace(raw_name[0], raw_name[0].lower(), 1)
    return name.replace("_", " ").strip()


def is_parser_function(template_name: str) -> bool:
    if template_name != "" and template_name.split(":")[0] in parser_functions:
        return True
    return False



from collections import Counter
from dataclasses import dataclass
import re
import unicodedata

import httpx
import mwparserfromhell as mwp
import pypandoc
from wikipediaapi import Wikipedia

from goodwiki.constants import removed_sections, templates_to_keep, country_templates
from goodwiki.errors import CorruptWikitextError
from goodwiki.mwp_utils import (
    try_remove_obj,
    is_media_wikilink,
    RE_CLEAN_WIKILINK,
    RE_RM_MAGIC,
    is_category_wikilink,
    clean_wikilink,
    get_heading,
    try_replace_obj,
    parse_params,
    get_wikilink_text,
    get_template_name,
    is_parser_function,
)
from goodwiki.tags import TagTranscluder


@dataclass
class RawWikiPage:
    wikitext: str
    pageid: int
    revid: int


@dataclass
class PreprocessedWikitext:
    wikitext: str
    categories: list[str]
    unsupported_templates: Counter[str]
    description: str | None


@dataclass
class WikiPage:
    pageid: int
    title: str
    lang: str
    revid: int
    categories: list[str]
    description: str | None
    raw_wikitext: str
    wikitext: str
    markdown: str
    unsupported_tags: Counter[str]
    unsupported_templates: Counter[str]


WIKI_API_ROOT = "https://en.wikipedia.org/w/api.php"


class GoodwikiClient:
    markdown_list_double_space_regex = re.compile(r"^([0-9]+\.|-)(  +)", re.MULTILINE)
    markdown_empty_unordered_list_item_regex = re.compile(r"^-( *)$\n", re.MULTILINE)

    def __init__(self, user_agent: str = "goodwiki/1.0 (https://euirim.org)"):
        self.user_agent = user_agent

    def get_category_pages(self, category: str) -> list[str]:
        """
        Get articles in Wikipedia's main namespace belonging to the given category.

        Args:
            category: name of the Wikipedia category; include "Category:" prefix.

        Returns:
            list of page titles associated with given category.
        """
        wikiapi = Wikipedia(self.user_agent, "en")
        cat = wikiapi.page(category)
        # 0 namespace is normal articles on wikipedia (non-category, etc.)
        return [c.title for c in cat.categorymembers.values() if c.namespace == 0]

    async def get_page_from_wikitext(
        self, raw_wikitext: str, title: str, pageid: int, revid: int, with_styling: bool
    ):
        """
        Get WikiPage object from given wikitext.

        Args:
            raw_wikitext: unprocessed wikitext of page.
            title: title of page.
            pageid: page id of page.
            revid: revision id of page.
            with_styling: whether to include styling syntax in the markdown like bolding, italicizing.

        Returns:
            WikiPage object.
        """
        # Preprocess wikitext
        preprocessed = self.preprocess_wikitext(raw_wikitext)

        # Expand wikitext templates
        expanded_wikitext = await self.expand_templates(preprocessed.wikitext)

        # Postprocess to clean up html
        final_wikitext, unsupported_tags = self.postprocess_wikitext(
            expanded_wikitext, with_styling
        )

        # Convert to markdown
        markdown = self.wikitext_to_markdown(final_wikitext)

        return WikiPage(
            pageid=pageid,
            title=unicodedata.normalize("NFKC", title),
            lang="en",
            revid=revid,
            categories=sorted(preprocessed.categories),
            description=preprocessed.description,
            raw_wikitext=raw_wikitext,
            wikitext=final_wikitext,
            markdown=markdown,
            unsupported_tags=unsupported_tags,
            unsupported_templates=preprocessed.unsupported_templates,
        )

    async def get_page(self, title: str, with_styling: bool = False) -> WikiPage:
        """
        Download Wikipedia page with the given title. Parse it to get markdown
        and meta information like categories all packaged in a WikiPage
        object.

        Args:
            title: title of the Wikipedia page.
            with_styling: whether to include styling syntax in the markdown like bolding, italicizing.

        Returns:
            WikiPage object.

        Raises:
            CorruptWikitextError: if the downloaded wikitext is corrupt/invalid.
        """
        # Download raw page
        raw_page = await self.download_raw_page(title)

        return await self.get_page_from_wikitext(
            raw_page.wikitext,
            title=title,
            pageid=raw_page.pageid,
            revid=raw_page.revid,
            with_styling=with_styling,
        )

    async def download_raw_page(self, title: str) -> RawWikiPage:
        transport = httpx.AsyncHTTPTransport(retries=1)
        async with httpx.AsyncClient(transport=transport, timeout=30.0) as client:
            res = await client.post(
                f"{WIKI_API_ROOT}",
                data={
                    "action": "query",
                    "prop": "revisions",
                    "titles": title,
                    "rvslots": "main",
                    "rvprop": "ids|content",
                    "format": "json",
                },
                headers={
                    "User-Agent": self.user_agent,
                },
            )
        assert res.status_code == 200
        res = res.json()["query"]
        page = res["pages"][list(res["pages"].keys())[0]]
        pageid = page["pageid"]
        rev = page["revisions"][0]
        content = rev["slots"]["main"]["*"]
        revid = rev["revid"]
        return RawWikiPage(wikitext=content, pageid=pageid, revid=revid)

    def _handle_section_removal(self, section) -> None:
        cur_section = section
        while cur_section is not None:
            heading = get_heading(section)
            if heading in removed_sections:
                continue

    def preprocess_wikitext(self, wikitext: str) -> PreprocessedWikitext:
        """
        Preprocess wikitext in preparation for expanding templates, also
        extracting a list of categories in the process.

        Do minimal processing here as the wikitext structure may be abnormal
        as template transclusion may be required to form properly structured wikitext.

        As a result, parsing in this function may be unreliable (e.g. getting sections).
        """
        code = mwp.parse(wikitext, skip_style_tags=True)

        first_pass_sections = []
        categories = []
        for section in code.get_sections(
            flat=True, include_lead=True, include_headings=True
        ):
            # Remove comments
            for comment in section.ifilter_comments():
                try_remove_obj(comment, section)

            # Get category links and parse wikilinks
            for obj in section.filter_wikilinks(recursive=True)[::-1]:
                if is_media_wikilink(obj):
                    try_remove_obj(obj, section)
                elif is_category_wikilink(obj):
                    categories.append(
                        unicodedata.normalize(
                            "NFKC",
                            re.sub(RE_CLEAN_WIKILINK, "", str(obj.title)).strip(),
                        )
                    )
                    clean_wikilink(obj)
                else:
                    try_replace_obj(obj, get_wikilink_text(obj), section)

            # Apply magic code (!, table endings) and remove known big templates
            # This may seem redundant but should minimize postprocessing errors
            # after expanding templates (like corrupt wikitext/html, etc.)
            for temp in section.filter_templates(recursive=True)[::-1]:
                temp_name = get_template_name(temp)
                if temp_name == "!":
                    try_replace_obj(temp, "|", section)
                elif temp_name in {"end", "jctbtm"}:
                    try_replace_obj(temp, "|}", section)

            # Remove behavior switches (parser instructions)
            # See: https://en.wikipedia.org/wiki/Help:Magic_words
            first_pass_sections.append(re.sub(RE_RM_MAGIC, "", str(section)))

        first_pass_wikitext = "\n\n".join(first_pass_sections)
        code = mwp.parse(first_pass_wikitext)

        preprocessed_sections = []
        unsupported_templates = []
        short_description = None
        for section in code.get_sections(
            levels=[2], include_lead=True, include_headings=True
        ):
            # Most templates should be removed, since they are asides to the main
            # content. But some templates should be kept to preserve content
            # (like smallcaps, strike)
            for temp in section.filter_templates(recursive=True)[::-1]:
                temp_name = get_template_name(temp)
                if (
                    temp_name not in templates_to_keep
                    and temp_name not in country_templates
                    and not temp_name.startswith("lang-")
                    and not is_parser_function(temp_name)
                ):
                    try_remove_obj(temp, section)
                    if temp_name == "short description":
                        params = parse_params(temp)
                        short_description = unicodedata.normalize(
                            "NFKC", params.get("1", "").strip()
                        )
                        if short_description in ["none", ""]:
                            short_description = None
                    else:
                        unsupported_templates.append(temp_name)

            # Don't strip whitespace from ends of section due to it causing
            # issues with lists
            preprocessed_sections.append(str(section))

        return PreprocessedWikitext(
            wikitext="\n\n".join(preprocessed_sections),
            categories=categories,
            unsupported_templates=Counter(unsupported_templates),
            description=short_description.strip()
            if short_description is not None
            else None,
        )

    async def expand_templates(self, wikitext: str) -> str:
        transport = httpx.AsyncHTTPTransport(retries=1)
        async with httpx.AsyncClient(transport=transport, timeout=30.0) as client:
            res = await client.post(
                f"{WIKI_API_ROOT}",
                data={
                    "action": "expandtemplates",
                    "prop": "wikitext",
                    "text": wikitext,
                    "format": "json",
                },
                headers={
                    "User-Agent": self.user_agent,
                },
            )
        assert res.status_code == 200
        return res.json()["expandtemplates"]["wikitext"].strip()

    def postprocess_wikitext(
        self, wikitext: str, with_styling: bool
    ) -> tuple[str, Counter[str]]:
        """
        Postprocess wikitext after expanding templates in preparation for
        markdown conversion.
        """
        # need to skip style tags to avoid some rare section parsing errors
        code = mwp.parse(wikitext, skip_style_tags=True)

        # Remove undesirable sections (e.g. references)
        # Needs to be separate to handle subsections gracefully
        for section in code.get_sections(include_lead=True, include_headings=True)[
            ::-1
        ]:
            heading = get_heading(section)
            if heading in removed_sections:
                code.remove(section)
                continue

        # Reassemble to read style tags
        code = mwp.parse(str(code), skip_style_tags=False)

        sections = []
        unsupported_tags = []
        transcluder = TagTranscluder(with_styling)
        for section in code.get_sections(
            levels=[2], include_lead=True, include_headings=True
        ):
            # Remove comments
            for comment in section.ifilter_comments():
                try_remove_obj(comment, section)

            # Transclude tags
            for tag in section.filter_tags(recursive=True)[::-1]:
                trans_out = transcluder.transclude(tag, section)
                if trans_out.unknown:
                    unsupported_tags.append(trans_out.tag_name)

            # Remove wikilinks in favor of text; remove category links outright
            # if they don't have text. Remove media links outright
            for obj in section.filter_wikilinks(recursive=True)[::-1]:
                if is_media_wikilink(obj):
                    try_remove_obj(obj, section)
                elif is_category_wikilink(obj):
                    if obj.text is None:
                        try_remove_obj(obj, section)
                    else:
                        clean_wikilink(obj)
                else:
                    try_replace_obj(obj, get_wikilink_text(obj), section)

            # Remove external links
            for obj in section.filter_external_links(recursive=True)[::-1]:
                if obj.title is None or obj.title == "":
                    try_remove_obj(obj, section)
                    continue
                try_replace_obj(obj, obj.title, section)

            sections.append(str(section))

        # Be sure doc ends with newline
        return "\n\n".join(sections).strip() + "\n", Counter(unsupported_tags)

    def wikitext_to_markdown(self, wikitext: str) -> str:
        markdown = ""
        try:
            markdown = pypandoc.convert_text(
                wikitext,
                "gfm",
                format="mediawiki",
                extra_args=["--wrap=none"],
            )
        except RuntimeError as err:
            raise CorruptWikitextError from err

        # Normalize unicode using NFKC (removing characters like non-breaking spaces)
        markdown = unicodedata.normalize("NFKC", markdown)

        # Remove errant comments in markdown (empty spaces)
        markdown = re.sub(r"<!--.*?-->", "", markdown)
        # Sometimes tables leave weird artifacts
        markdown = re.sub(r"\\\|\}", "", markdown)
        # Remove lines with just spaces (should be after removing comments due
        # to some comments taking up full line)
        # TODO: This may cause code to be compressed
        markdown = re.sub(r"\n\s*\n", "\n\n", markdown)
        # Remove spaces before newline (trailing spaces)
        markdown = re.sub(r" *\n", "\n", markdown)
        # Remove double spaces in front of list items (pandoc problem)
        markdown = re.sub(self.markdown_list_double_space_regex, r"\1 ", markdown)
        # Remove empty unordered list items
        markdown = re.sub(self.markdown_empty_unordered_list_item_regex, "", markdown)
        # Avoid excessive escaping \[]
        # \$, \# continues to be escaped due to conflicts with math
        markdown = re.sub(r"\\(\[|\])", r"\1", markdown)

        # Remove errors/artifacts
        markdown = re.sub(
            r'Format price error: cannot parse value "(.*?)"', r"\1", markdown
        )
        markdown = markdown.replace("<sup>[update]</sup>", "")

        return markdown.strip()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['goodwiki']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.6,<9.0.0',
 'httpx>=0.24.1,<0.25.0',
 'multiprocess>=0.70.15,<0.71.0',
 'mwparserfromhell>=0.6.4,<0.7.0',
 'pyarrow>=12.0.1,<13.0.0',
 'pypandoc>=1.11,<2.0',
 'tqdm>=4.66.1,<5.0.0',
 'wikipedia-api>=0.6.0,<0.7.0']

setup_kwargs = {
    'name': 'goodwiki',
    'version': '1.0.1',
    'description': 'Utility that converts Wikipedia pages into GitHub-flavored Markdown.',
    'long_description': '# GoodWiki\n\nGoodWiki is a Python package that carefully converts Wikipedia pages into GitHub-flavored Markdown. Converted pages preserve layout features like lists, code blocks, math, and block quotes.\n\nThis package is used to generate the [GoodWiki Dataset](https://github.com/euirim/goodwiki).\n\n## Installation\n\nThis package supports Python 3.11+.\n\n1. Install via pip.\n\n```bash\npip install goodwiki\n```\n\n2. Install pandoc v2.19.2. Follow instructions [here](https://pandoc.org/installing.html).\n\n## Usage\n\n### Initializing Client\n\n```python\nimport asyncio\nfrom goodwiki import GoodwikiClient\n\nclient = GoodwikiClient()\n```\n\nYou can also optionally provide your own user agent (default is `goodwiki/1.0 (https://euirim.org)`):\n\n```python\n\nclient = GoodwikiClient("goodwiki/1.0 (bob@gmail.com)")\n```\n\n### Getting Single Page\n\n```python\npage = asyncio.run(client.get_page("Usain Bolt"))\n```\n\nYou can also optionally include styling syntax like bolding to the final markdown:\n\n```python\npage = asyncio.run(client.get_page("Usain Bolt", with_styling=True))\n```\n\nYou can access the resulting data via properties. For example:\n\n```python\nprint(page.markdown)\n```\n\n### Getting Category Pages\n\nTo get a list of page titles associated with a Wikipedia category, run the following:\n\n```python\nclient.get_category_pages("Category:Good_articles")\n```\n\n### Converting Existing Raw Wikitext\n\nIf you\'ve already downloaded raw wikitext from Wikipedia, you can convert it to Markdown by running:\n\n```python\nclient.get_page_from_wikitext(\n\traw_wikitext="RAW_WIKITEXT",\n\t# The rest of the fields are meant for populating the final WikiPage object\n\ttitle="Usain Bolt",\n\tpageid=123,\n\trevid=123,\n)\n```\n\n## Methodology\n\nFull details are available in this package\'s [GitHub repo README](https://github.com/euirim/goodwiki).\n\n## External Links\n\n* [Changelog](https://github.com/euirim/goodwiki/releases)\n* [GitHub](https://github.com/euirim/goodwiki)\n* [Dataset](https://huggingface.co/datasets/euirim/goodwiki)\n',
    'author': 'Euirim Choi',
    'author_email': 'euirim@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/euirim/goodwiki',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)
